"""Miscellaneous utility functions.
"""
import sys
from math import floor, log10


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def pretty_num(x, sig=4):
    return str('%g' % (round_sig(x, sig)))


def round_sig(x, sig=4):
    if x == 0.0:
        return 0.0
    return round(x, sig - int(floor(log10(abs(x)))) - 1)


def listify(x):
    if not isinstance(x, list):
        return [x]
    return x


def print_inline(x, new_line=False):
    if not new_line:
        sys.stdout.write("\033[F")
        sys.stdout.write("\033[K")
    print(x, flush=True)



from ..module import Module

CLASS_NAME = 'LightCurve'


class LightCurve(Module):
    """Output a light curve to disk.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._n_times = kwargs.get('ntimes', 0)

    def process(self, **kwargs):
        output = {}
        for key in ['magnitudes', 'e_magnitudes', 'model_magnitudes',
                    'bands', 'times']:
            output[key] = kwargs[key]

        return output



import numpy as np

from ...utils import is_number, listify
from ..module import Module

CLASS_NAME = 'Transient'


class Transient(Module):
    """Structure to store transient data.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._keys = kwargs.get('keys', '')

    def process(self, **kwargs):
        return self._data

    def set_data(self, all_data, req_key_values={}, subtract_minimum_keys=[]):
        self._all_data = all_data
        self._data = {}
        if not self._all_data:
            return
        name = list(self._all_data.keys())[0]
        for key in self._keys:
            if key not in self._all_data[name]:
                continue
            subdata = self._all_data[name][key]
            subkeys = self._keys[key]
            req_subkeys = [
                x for x in subkeys
                if not isinstance(subkeys, dict) or 'required' in listify(
                    subkeys[x])
            ]
            num_subkeys = [
                x for x in subkeys if 'numeric' in listify(subkeys[x])
            ]
            exc_subkeys = [
                x for x in subkeys if 'exclude' in listify(subkeys[x])
            ]
            # Only include data that contains all subkeys
            for entry in subdata:
                if any([x not in entry for x in req_subkeys]):
                    continue
                if any([x in entry for x in exc_subkeys]):
                    continue
                if any([not is_number(entry[x]) or np.isnan(float(entry[x]))
                        for x in num_subkeys]):
                    continue
                skip_key = False
                for qkey in req_key_values:
                    if (qkey in entry and
                            entry[qkey] not in req_key_values[qkey]):
                        skip_key = True
                        break
                if skip_key:
                    continue
                for x in subkeys:
                    if x == 'value':
                        self._data[key] = entry.get(x, '')
                    else:
                        self._data.setdefault(x + 's',
                                              []).append(entry.get(x, ''))

        for key in self._data.copy():
            if isinstance(self._data[key], list):
                if not all(is_number(x) for x in self._data[key]):
                    continue
                self._data[key] = [float(x) for x in self._data[key]]
                self._data['min_' + key] = min(self._data[key])
                self._data['max_' + key] = max(self._data[key])
            else:
                if is_number(self._data[key]):
                    self._data[key] = float(self._data[key])

        for qkey in subtract_minimum_keys:
            minv = self._data['min_' + qkey]
            self._data[qkey] = [x - minv for x in self._data[qkey]]



import datetime
import importlib
import json
import logging
import os
import shutil
import time
from collections import OrderedDict
from math import isnan

import emcee
import numpy as np
from emcee.utils import MPIPool
from scipy.optimize import basinhopping

from .utils import listify, pretty_num, print_inline, round_sig


class Model:
    """Define a semi-analytical model to fit transients with.
    """

    MODEL_OUTPUT_DIR = 'mosfit-products'

    def __init__(self,
                 parameter_path='parameters.json',
                 model='default',
                 travis=False):
        self._model_name = model
        self._travis = travis

        self._dir_path = os.path.dirname(os.path.realpath(__file__))

        # Load the model file.
        model = self._model_name
        model_dir = self._model_name

        if '.json' in self._model_name:
            model_dir = self._model_name.split('.json')[0]
        else:
            model = self._model_name + '.json'

        model_path = os.path.join(self._dir_path, 'models', model_dir, model)

        if os.path.isfile(model):
            model_path = model

        print('Model file: ' + model_path)

        with open(model_path, 'r') as f:
            self._model = json.loads(f.read())

        # Load model parameter file.
        model_pp = os.path.join(
            os.path.split(model_path)[0], 'parameters.json')

        pp = model_pp

        # First try user-specified path
        if parameter_path and os.path.isfile(parameter_path):
            pp = parameter_path
        # Then try directory we are running from
        elif os.path.isfile('parameter.json'):
            pp = 'parameter.json'
        # Finally try model folder
        elif os.path.isfile(model_pp):
            pp = model_pp

        print('Parameter file: ' + pp + '\n')

        with open(pp, 'r') as f:
            self._parameters = json.loads(f.read())
        self._num_free_parameters = len(
            [x for x in self._parameters
             if ('min_value' in self._parameters[x] and 'max_value' in
                 self._parameters[x])])
        self._log = logging.getLogger()
        self._modules = {}
        self._bands = []

        # Load the call tree for the model. Work our way in reverse from the
        # observables, first constructing a tree for each observable and then
        # combining trees.
        root_kinds = ['output', 'objective']

        trees = {}
        self.construct_trees(self._model, trees, kinds=root_kinds)

        unsorted_call_stack = {}
        self._max_depth_all = -1
        for tag in self._model:
            cur_model = self._model[tag]
            roots = []
            if cur_model['kind'] in root_kinds:
                max_depth = 0
                roots = [cur_model['kind']]
            else:
                max_depth = -1
                for tag2 in trees:
                    roots.extend(trees[tag2]['roots'])
                    depth = self.get_max_depth(tag, trees[tag2], max_depth)
                    if depth > max_depth:
                        max_depth = depth
                    if depth > self._max_depth_all:
                        self._max_depth_all = depth
            roots = list(set(roots))
            new_entry = cur_model.copy()
            new_entry['roots'] = roots
            if 'children' in new_entry:
                del (new_entry['children'])
            new_entry['depth'] = max_depth
            unsorted_call_stack[tag] = new_entry
        # print(unsorted_call_stack)

        # Currently just have one call stack for all products, can be wasteful
        # if only using some products.
        self._call_stack = OrderedDict()
        for depth in range(self._max_depth_all, -1, -1):
            for task in unsorted_call_stack:
                if unsorted_call_stack[task]['depth'] == depth:
                    self._call_stack[task] = unsorted_call_stack[task]

        for task in self._call_stack:
            cur_task = self._call_stack[task]
            class_name = cur_task.get('class', task)
            mod = importlib.import_module(
                '.' + 'modules.' + cur_task['kind'] + 's.' + class_name,
                package='mosfit')
            mod_class = getattr(mod, mod.CLASS_NAME)
            if cur_task['kind'] == 'parameter' and task in self._parameters:
                cur_task.update(self._parameters[task])
            self._modules[task] = mod_class(name=task, **cur_task)
            if class_name == 'filters':
                self._bands = self._modules[task].band_names()

        for task in reversed(self._call_stack):
            cur_task = self._call_stack[task]
            if 'requests' in cur_task:
                inputs = listify(cur_task.get('inputs', []))
                for i, inp in enumerate(inputs):
                    requests = {}
                    parent = ''
                    for par in self._call_stack:
                        if par == inp:
                            parent = par
                    if not parent:
                        self._log.error(
                            "Couldn't find parent task for  {}!".format(inp))
                        raise ValueError
                    reqs = cur_task['requests'][i]
                    for req in reqs:
                        requests[req] = self._modules[task].request(req)
                    self._modules[parent].handle_requests(**requests)

    def basinhop(self, x):
        """Perform Basin-hopping upon a single walker.
        """
        bh = basinhopping(self.bhprob, x, niter=10)
        return bh

    def construct_trees(self, d, trees, kinds=[], name='', roots=[], depth=0):
        """Construct call trees for each root.
        """
        for tag in d:
            entry = d[tag].copy()
            new_roots = roots
            if entry['kind'] in kinds or tag == name:
                entry['depth'] = depth
                if entry['kind'] in kinds:
                    new_roots.append(entry['kind'])
                entry['roots'] = list(set(new_roots))
                trees[tag] = entry
                inputs = listify(entry.get('inputs', []))
                for inp in inputs:
                    children = {}
                    self.construct_trees(
                        d,
                        children,
                        name=inp,
                        roots=new_roots,
                        depth=depth + 1)
                    trees[tag].setdefault('children', {})
                    trees[tag]['children'].update(children)

    def draw_walker(self, test=True):
        """Draw a walker randomly from the full range of all parameters, reject
        walkers that return invalid scores.
        """
        p = None
        while p is None:
            draw = np.random.uniform(low=0.0, high=1.0, size=self._n_dim)
            if not test:
                p = draw
                break
            score = self.likelihood(draw)
            if not isnan(score) and np.isfinite(score):
                p = draw
        return p

    def fit_data(self,
                 data,
                 event_name='',
                 plot_points=[],
                 iterations=2000,
                 frack_step=100,
                 num_walkers=100,
                 num_temps=2,
                 fracking=True,
                 post_burn=500):
        """Fit the data for a given event with this model using a combination
        of emcee and Basin-hopping.
        """
        for task in self._call_stack:
            cur_task = self._call_stack[task]
            self._modules[task].set_event_name(event_name)
            if cur_task['kind'] == 'data':
                self._modules[task].set_data(
                    data,
                    req_key_values={'band': self._bands},
                    subtract_minimum_keys=['times'])

        ntemps, ndim, nwalkers = (num_temps, self._num_free_parameters,
                                  num_walkers)
        self._event_name = event_name
        self._n_dim = ndim
        self._emcee_est_t = 0.0
        self._bh_est_t = 0.0
        self._fracking = fracking
        self._burn_in = max(iterations - post_burn, 0)

        test_walker = iterations > 0
        lnprob = None
        sampler_args = {}
        serial = False
        pool = ''
        try:
            pool = MPIPool()
        except ValueError:
            psize = 1
            serial = True
        except:
            raise
        else:
            sampler_args = {'pool': pool}
            psize = pool.size

        if serial or pool.is_master():
            print_inline('{} dimensions in problem.'.format(ndim))
            p0 = [[] for x in range(ntemps)]

            for i, pt in enumerate(p0):
                while len(p0[i]) < nwalkers:
                    self.print_status(
                        desc='Drawing initial walkers',
                        progress=[i * nwalkers + len(p0[i]),
                                  nwalkers * ntemps])
                    if serial:
                        p0[i].append(self.draw_walker(test_walker))
                    else:
                        nmap = nwalkers - len(p0[i])
                        p0[i].extend(
                            pool.map(self.draw_walker, [test_walker] * nmap))
                # p0[i].extend(pool.map(self.draw_walker, range(nwalkers)))
        else:
            pool.wait()
            return

        sampler = emcee.PTSampler(ntemps, nwalkers, ndim, self.likelihood,
                                  self.prior, **sampler_args)

        print_inline('Initial draws completed!\n')
        p = p0.copy()

        if fracking:
            frack_iters = max(round(iterations / frack_step), 1)
            bmax = int(round(self._burn_in / float(frack_step)))
            loop_step = frack_step
        else:
            frack_iters = 1
            loop_step = iterations

        for b in range(frack_iters):
            if fracking and b >= bmax:
                loop_step = iterations - self._burn_in
            emi = 0
            st = time.time()
            for p, lnprob, lnlike in sampler.sample(
                    p, iterations=min(loop_step, iterations)):
                emi = emi + 1
                self._emcee_est_t = float(time.time() - st) / emi * (
                    iterations - (b * frack_step + emi))
                self.print_status(
                    desc='Running PTSampler',
                    scores=[max(x) for x in lnprob],
                    progress=[b * frack_step + emi, iterations])
            if fracking and b >= bmax:
                break
            if fracking and b < bmax:
                self.print_status(
                    desc='Running Basin-hopping',
                    scores=[max(x) for x in lnprob],
                    progress=[(b + 1) * frack_step, iterations])
                ris, rjs = [0] * psize, np.random.randint(nwalkers, size=psize)

                bhwalkers = [p[i][j] for i, j in zip(ris, rjs)]
                st = time.time()
                if serial:
                    bhs = list(map(self.basinhop, bhwalkers))
                else:
                    bhs = pool.map(self.basinhop, bhwalkers)
                for bhi, bh in enumerate(bhs):
                    p[ris[bhi]][rjs[bhi]] = bh.x
                self._bh_est_t = float(time.time() - st) * (bmax - b - 1)
                scores = [-x.fun for x in bhs]
                self.print_status(
                    desc='Running Basin-hopping',
                    scores=scores,
                    progress=[(b + 1) * frack_step, iterations])

        if pool:
            pool.close()

        walkers_out = OrderedDict()
        for xi, x in enumerate(p[0]):
            walkers_out[xi] = self.run_stack(x, root='output')
            if lnprob is not None:
                walkers_out[xi]['score'] = lnprob[0][xi]
            parameters = {}
            pi = 0
            for ti, task in enumerate(self._call_stack):
                cur_task = self._call_stack[task]
                if (cur_task['kind'] != 'parameter' or
                        'min_value' not in cur_task or
                        'max_value' not in cur_task):
                    continue
                output = self._modules[task].process(**{'fraction': x[pi]})
                value = list(output.values())[0]
                paramdict = {
                    'value': value,
                    'fraction': x[pi],
                    'latex': self._modules[task].latex(),
                    'log': self._modules[task].is_log()
                }
                parameters.update({self._modules[task].name(): paramdict})
                pi = pi + 1
            walkers_out[xi]['parameters'] = parameters

        if not os.path.isfile('mosfit.ipynb'):
            shutil.copy(
                os.path.join(self._dir_path, 'jupyter', 'mosfit.ipynb'),
                os.path.join(os.getcwd(), 'mosfit.ipynb'))

        if not os.path.exists(self.MODEL_OUTPUT_DIR):
            os.makedirs(self.MODEL_OUTPUT_DIR)

        with open(os.path.join(self.MODEL_OUTPUT_DIR, 'walkers.json'),
                  'w') as flast, open(
                      os.path.join(self.MODEL_OUTPUT_DIR,
                                   self._event_name + '.json'), 'w') as f:
            json.dump(walkers_out, flast, indent='\t', separators=(',', ':'))
            json.dump(walkers_out, f, indent='\t', separators=(',', ':'))

        return (p, lnprob)

    def print_status(self, desc='', scores='', progress=''):
        """Prints a status message showing the current state of the fitting process.
        """
        outarr = [self._event_name]
        if desc:
            outarr.append(desc)
        if isinstance(scores, list):
            scorestring = 'Best scores: [ ' + ', '.join(
                [pretty_num(x) for x in scores]) + ' ]'
            outarr.append(scorestring)
        if isinstance(progress, list):
            progressstring = 'Progress: [ {}/{} ]'.format(*progress)
            outarr.append(progressstring)
        if self._emcee_est_t + self._bh_est_t > 0.0:
            if self._bh_est_t > 0.0 or not self._fracking:
                tott = self._emcee_est_t + self._bh_est_t
            else:
                tott = 2.0 * self._emcee_est_t
            timestring = self.get_timestring(tott)
            outarr.append(timestring)

        print_inline(' | '.join(outarr), new_line=self._travis)

    def get_timestring(self, t):
        """Return a string showing the estimated remaining time based upon
        elapsed times for emcee and Basin-hopping.
        """
        return ('Estimated time left: [ ' + str(
            datetime.timedelta(seconds=(round_sig(t)))).rstrip('0').rstrip('.')
                + ' ]')

    def get_max_depth(self, tag, parent, max_depth):
        """Return the maximum depth a given task is found in a tree.
        """
        for child in parent.get('children', []):
            if child == tag:
                new_max = parent['children'][child]['depth']
                if new_max > max_depth:
                    max_depth = new_max
            else:
                new_max = self.get_max_depth(tag, parent['children'][child],
                                             max_depth)
                if new_max > max_depth:
                    max_depth = new_max
        return max_depth

    def likelihood(self, x):
        """Return score related to maximum likelihood.
        """
        outputs = self.run_stack(x, root='objective')
        return outputs['value']

    def prior(self, data):
        """Return score related to paramater priors.
        """
        return 0.0

    def bhprob(self, x):
        """Return score for basinhopping.
        """
        return -self.likelihood(x)

    def run_stack(self, x, root='objective'):
        """Run a stack of modules as defined in the model definition file. Only
        run functions that match the specified root.
        """
        inputs = {}
        outputs = {}
        pos = 0
        cur_depth = self._max_depth_all
        for task in self._call_stack:
            cur_task = self._call_stack[task]
            if root not in cur_task['roots']:
                continue
            if cur_task['depth'] != cur_depth:
                inputs = outputs
            cur_depth = cur_task['depth']
            if (cur_task['kind'] == 'parameter' and 'min_value' in cur_task and
                    'max_value' in cur_task):
                inputs.update({'fraction': x[pos]})
                inputs.setdefault('fractions', []).append(x[pos])
                pos = pos + 1
            new_outs = self._modules[task].process(**inputs)
            outputs.update(new_outs)

            if cur_task['kind'] == root:
                return outputs



import numpy as np
from astropy import constants as c
from astropy import units as u

from ..module import Module

CLASS_NAME = 'SED'


class SED(Module):
    """Template class for SED Modules.
    """

    C_CONST = (c.c / u.Angstrom).cgs.value
    N_PTS = 16 + 1

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._band_wavelengths = []
        self._filters = []

    def handle_requests(self, **requests):
        self._filters = requests.get('filters', [])
        self._band_wavelengths = requests.get('bandwavelengths', [])
        if not self._band_wavelengths:
            wave_ranges = requests.get('band_wave_ranges', [])
            if not wave_ranges:
                return
            for rng in wave_ranges:
                self._band_wavelengths.append(
                    list(np.linspace(rng[0], rng[1], self.N_PTS)))
        self._band_frequencies = [[self.C_CONST / x for x in y]
                                  for y in self._band_wavelengths]

    def add_to_existing_seds(self, new_seds, **kwargs):
        old_seds = kwargs.get('seds', None)
        if old_seds is not None:
            new_seds = [(i + j for i, j in zip(x, y))
                        for x, y in zip(old_seds, list(new_seds))]
        return new_seds

    def request(self, request):
        if request == 'filters':
            return self._filters
        elif request == 'bandwavelengths':
            return self._band_wavelengths
        return []



"""List of numerical constants, faster than using astropy with every call.
"""

import astropy.constants as c
import astropy.units as u
import numpy as np

LOG2 = np.log(2.0)
LIKELIHOOD_FLOOR = -np.inf
FOUR_PI = 4.0 * np.pi
MAG_FAC = 100.0**0.2
AB_OFFSET = -48.60
MPC_CGS = (1.0 * u.Mpc).cgs.value
DAY_CGS = (1.0 * u.day).cgs.value
M_SUN_CGS = c.M_sun.cgs.value
KM_CGS = 1.0e5
C_CGS = c.c.cgs.value



from math import isnan

import numpy as np

from .engine import Engine

CLASS_NAME = 'NickelCobalt'


class NickelCobalt(Engine):
    """Nickel/Cobalt decay engine
    """

    NI56_LUM = 6.45e43
    CO56_LUM = 1.45e43
    NI56_LIFE = 8.8
    CO56_LIFE = 111.3

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process(self, **kwargs):
        if 'densetimes' in kwargs:
            self._times = kwargs['densetimes']
        else:
            self._times = kwargs['times']
        self._mnickel = kwargs['fnickel'] * kwargs['mejecta']
        self._texplosion = kwargs['texplosion']

        # From 1994ApJS...92..527N
        ts = [np.inf if self._texplosion > x else (x - self._texplosion)
              for x in self._times]
        luminosities = [self._mnickel * (self.NI56_LUM * np.exp(
            -t / self.NI56_LIFE) + self.CO56_LUM * np.exp(-t / self.CO56_LIFE))
                        for t in ts]
        luminosities = [0.0 if isnan(x) else x for x in luminosities]

        # Add on to any existing luminosity
        old_luminosities = kwargs.get('luminosities', None)
        if old_luminosities is not None:
            luminosities = [x + y
                            for x, y in zip(old_luminosities, luminosities)]

        # Add on to any existing luminosity
        luminosities = self.add_to_existing_lums(luminosities)

        return {'luminosities': luminosities}



class Module:
    def __init__(self, name, **kwargs):
        self._name = name
        self._log = False

    def process(self, **kwargs):
        return {}

    def request(self, request):
        return []

    def name(self):
        return self._name

    def is_log(self):
        return self._log

    def handle_requests(self, **kwargs):
        pass

    def set_event_name(self, event_name):
        self._event_name = event_name



from math import pi

import numexpr as ne
import numpy as np
from astropy import constants as c

from ...constants import DAY_CGS, FOUR_PI, KM_CGS
from .sed import SED

CLASS_NAME = 'Photosphere'


class Photosphere(SED):
    """Expanding/recending photosphere with a blackbody spectral energy
    distribution.
    """

    FLUX_CONST = FOUR_PI * (2.0 * c.h / (c.c**2) * pi).cgs.value
    X_CONST = (c.h / c.k_B).cgs.value
    STEF_CONST = (4.0 * pi * c.sigma_sb).cgs.value
    RAD_CONST = KM_CGS * DAY_CGS

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process(self, **kwargs):
        self._t_explosion = kwargs['texplosion']
        self._times = kwargs['times']
        self._luminosities = kwargs['luminosities']
        self._temperature = kwargs['temperature']
        self._bands = kwargs['bands']
        self._v_ejecta = kwargs['vejecta']
        self._radius2 = [(self.RAD_CONST * self._v_ejecta *
                          (x - self._t_explosion))**2 for x in self._times]
        self._rec_radius2 = [x / (self.STEF_CONST * self._temperature**4)
                             for x in self._luminosities]
        xc = self.X_CONST
        fc = self.FLUX_CONST
        zp1 = 1.0 + kwargs['redshift']
        seds = []
        for li, lum in enumerate(self._luminosities):
            cur_band = self._bands[li]
            bi = self._filters.find_band_index(cur_band)
            rest_freqs = [x * zp1 for x in self._band_frequencies[bi]]

            # Radius is determined via expansion, unless this would make
            # temperature lower than temperature parameter.
            radius2 = self._radius2[li]
            rec_radius2 = self._rec_radius2[li]
            if radius2 < rec_radius2:
                temperature = (lum / (self.STEF_CONST * radius2))**0.25
            else:
                radius2 = rec_radius2
                temperature = self._temperature

            if li == 0:
                sed = ne.evaluate('fc * radius2 * rest_freqs**3 / '
                                  'exp(xc * rest_freqs / temperature) - 1.0')
            else:
                sed = ne.re_evaluate()
            seds.append(list(sed))

        seds = self.add_to_existing_seds(seds, **kwargs)

        return {'bandwavelengths': self._band_wavelengths, 'seds': seds}



import csv
import json
import os

import numexpr as ne
import numpy as np

from ...constants import AB_OFFSET, FOUR_PI, MAG_FAC, MPC_CGS
from ...utils import listify
from ..module import Module

CLASS_NAME = 'Filters'


class Filters(Module):
    """Band-pass filters.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        bands = kwargs.get('bands', '')
        systems = kwargs.get('systems', '')
        instruments = kwargs.get('instruments', '')
        bands = listify(bands)
        systems = listify(systems)
        instruments = listify(instruments)

        dir_path = os.path.dirname(os.path.realpath(__file__))
        band_list = []
        with open(
                os.path.join(dir_path, 'filterrules.json')) as f:
            filterrules = json.loads(f.read())
            for bi, band in enumerate(bands):
                for rule in filterrules:
                    if systems[bi] not in rule.get("systems", []):
                        continue
                    if instruments[bi] not in rule.get("instruments", []):
                        continue
                    for bnd in rule.get('filters', []):
                        if band == bnd or band == '':
                            band_list.append(rule['filters'][bnd])
                            band_list[-1]['systems'] = rule.get('systems', [])
                            band_list[-1]['instruments'] = rule.get(
                                'instruments', [])
                            band_list[-1]['name'] = bnd
                            if not band_list[-1].get('offset', ''):
                                band_list[-1]['offset'] = 0.0

        self._unique_bands = band_list
        self._band_insts = [x['instruments'] for x in self._unique_bands]
        self._band_systs = [x['systems'] for x in self._unique_bands]
        self._band_names = [x['name'] for x in self._unique_bands]
        self._band_offsets = [x['offset'] for x in self._unique_bands]
        self._n_bands = len(self._unique_bands)
        self._band_wavelengths = [[] for i in range(self._n_bands)]
        self._transmissions = [[] for i in range(self._n_bands)]
        self._min_waves = [0.0] * self._n_bands
        self._max_waves = [0.0] * self._n_bands
        self._filter_integrals = [0.0] * self._n_bands

        for i, band in enumerate(self._unique_bands):
            with open(
                    os.path.join(dir_path, 'filters', band['path']), 'r') as f:
                rows = []
                for row in csv.reader(f, delimiter=' ', skipinitialspace=True):
                    rows.append([float(x) for x in row[:2]])
            self._band_wavelengths[i], self._transmissions[i] = list(
                map(list, zip(*rows)))
            self._min_waves[i] = min(self._band_wavelengths[i])
            self._max_waves[i] = max(self._band_wavelengths[i])
            self._filter_integrals[i] = np.trapz(
                self._transmissions[i],
                self._band_wavelengths[i])

    def find_band_index(self, name, instrument='', system=''):
        for bi, band in enumerate(self._unique_bands):
            if (name == band['name'] and
                instrument in self._band_insts[bi] and
                    system in self._band_systs[bi]):
                return bi
            if (name == band['name'] and
                '' in self._band_insts[bi] and
                    '' in self._band_systs[bi]):
                return bi
        raise(ValueError('Cannot find band index!'))

    def process(self, **kwargs):
        self._dist_const = np.log10(FOUR_PI * (kwargs['lumdist'] * MPC_CGS)**2)
        self._luminosities = kwargs['luminosities']
        self._bands = kwargs['bands']
        self._systems = kwargs['systems']
        self._instruments = kwargs['instruments']
        eff_fluxes = []
        offsets = []
        for li, band in enumerate(self._luminosities):
            cur_band = self._bands[li]
            bi = self.find_band_index(cur_band, self._systems[li],
                                      self._instruments[li])
            sed = kwargs['seds'][li]
            wavs = kwargs['bandwavelengths'][bi]
            offsets.append(self._band_offsets[bi])
            dx = wavs[1] - wavs[0]
            itrans = np.interp(wavs, self._band_wavelengths[bi],
                               self._transmissions[bi])
            # if li == 0:
            #     ef = ne.evaluate('sum(itrans * sed)')
            # else:
            #     ef = ne.re_evaluate()
            # eff_fluxes.append(dx * ef)
            yvals = [x * y for x, y in zip(itrans, sed)]
            eff_fluxes.append(
                np.trapz(
                    yvals, dx=dx) / self._filter_integrals[bi])
        mags = self.abmag(eff_fluxes, offsets)
        return {'model_magnitudes': mags}

    def band_names(self):
        return self._band_names

    def abmag(self, eff_fluxes, offsets):
        return [(np.inf if x == 0.0 else
                 (AB_OFFSET - y - MAG_FAC * (np.log10(x) - self._dist_const)))
                for x, y in zip(eff_fluxes, offsets)]

    def request(self, request):
        if request == 'filters':
            return self
        elif request == 'band_wave_ranges':
            return list(map(list, zip(*[self._min_waves, self._max_waves])))
        return []



import numpy as np

from extinction import apply as eapp
from extinction import odonnell94

from .sed import SED

CLASS_NAME = 'Extinction'


class Extinction(SED):
    """Adds extinction to SED from both host galaxy and MW.
    """

    MW_RV = 3.1

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._preprocessed = False

    def process(self, **kwargs):
        self.preprocess(**kwargs)
        self._seds = kwargs['seds']
        self._nh_host = kwargs['nhhost']

        av_host = self._nh_host / 1.8e21

        for si in range(len(self._bands)):
            cur_band = self._bands[si]
            bi = self._filters.find_band_index(cur_band)
            # First extinct out LOS dust from MW
            eapp(self._mw_extinct[si], self._seds[si], inplace=True)
            # Then extinct out host gal (using rest wavelengths)
            eapp(
                odonnell94(
                    np.array(self._band_rest_wavelengths[bi]), av_host,
                    self.MW_RV),
                self._seds[si],
                inplace=True)

        return {'bandwavelengths': self._band_wavelengths, 'seds': self._seds}

    def preprocess(self, **kwargs):
        if not self._preprocessed:
            zp1 = 1.0 + kwargs['redshift']
            self._ebv = kwargs['ebv']
            self._bands = kwargs['bands']
            self._band_rest_wavelengths = [[y / zp1 for y in x]
                                           for x in self._band_wavelengths]
            self._av_mw = self.MW_RV * self._ebv
            self._mw_extinct = []
            for si in range(len(self._bands)):
                cur_band = self._bands[si]
                bi = self._filters.find_band_index(cur_band)
                # First extinct out LOS dust from MW
                self._mw_extinct.append(
                    odonnell94(
                        np.array(self._band_wavelengths[bi]), self._av_mw,
                        self.MW_RV))
        self._preprocessed = True



import json
import os
import warnings

import numpy as np
import requests
from emcee.utils import MPIPool

from .model import Model

warnings.filterwarnings("ignore")


class Fitter():
    """Fit transient events with the provided model.
    """

    def __init__(self):
        pass

    def fit_events(self,
                   events=[''],
                   models=[],
                   plot_points='',
                   max_time='',
                   band_list='',
                   band_systems='',
                   band_instruments='',
                   iterations=1000,
                   num_walkers=50,
                   num_temps=2,
                   parameter_paths=[],
                   fracking=True,
                   frack_step=20,
                   travis=False,
                   post_burn=500):
        dir_path = os.path.dirname(os.path.realpath(__file__))

        for event in events:
            if event:
                pool = ''
                try:
                    pool = MPIPool()
                except ValueError:
                    pass
                except:
                    raise
                event_name = ''
                if not pool or pool.is_master():
                    path = ''
                    # If the event name ends in .json, assume a path
                    if event.endswith('.json'):
                        path = event
                        event_name = event.replace('.json', '').split('/')[-1]
                    # If not (or the file doesn't exist), download from OSC
                    if not path or not os.path.exists(path):
                        names_path = os.path.join(dir_path, 'cache',
                                                  'names.min.json')
                        try:
                            response = requests.get(
                                'https://sne.space/astrocats/astrocats/'
                                'supernovae/output/names.min.json')
                        except:
                            print('Warning: Could not download SN names!')
                        else:
                            with open(names_path, 'wb') as f:
                                f.write(response.content)
                                f.flush()
                        if os.path.exists(names_path):
                            with open(names_path, 'r') as f:
                                names = json.loads(f.read())
                        else:
                            print('Error: Could not read list of SN names!')
                            raise RuntimeError

                        if event in names:
                            event_name = event
                        else:
                            for name in names:
                                if event in names[name]:
                                    event_name = name
                                    break
                        if not event_name:
                            print('Error: Could not find event by that name!')
                            raise RuntimeError
                        urlname = event_name + '.json'

                        name_path = os.path.join(dir_path, 'cache', urlname)
                        try:
                            response = requests.get(
                                'https://sne.space/astrocats/astrocats/'
                                'supernovae/output/json/' + urlname)
                        except:
                            print('Warning: Could not download SN data!')
                        else:
                            with open(name_path, 'wb') as f:
                                f.write(response.content)
                                f.flush()
                        path = name_path

                if pool:
                    if pool.is_master():
                        for rank in range(1, pool.size + 1):
                            pool.comm.send(event_name, dest=rank, tag=0)
                            pool.comm.send(path, dest=rank, tag=1)
                    else:
                        event_name = pool.comm.recv(source=0, tag=0)
                        path = pool.comm.recv(source=0, tag=1)
                        pool.wait()
                if pool:
                    pool.close()

                if os.path.exists(path):
                    with open(path, 'r') as f:
                        data = json.loads(f.read())
                else:
                    print('Error: Could not find supernova data locally or '
                          'on the OSC.')
                    raise RuntimeError

            for mod_name in models:
                for parameter_path in parameter_paths:
                    model = Model(
                        model=mod_name,
                        parameter_path=parameter_path,
                        travis=travis)

                    if not event:
                        print('No event specified, generating dummy data.')
                        event_name = mod_name
                        gen_args = {
                            'name': mod_name,
                            'max_time': max_time,
                            'plot_points': plot_points,
                            'band_list': band_list,
                            'band_systems': band_systems,
                            'band_instruments': band_instruments
                        }
                        data = self.generate_dummy_data(**gen_args)

                    model.fit_data(
                        data,
                        event_name=event_name,
                        plot_points=plot_points,
                        iterations=iterations,
                        num_walkers=num_walkers,
                        num_temps=num_temps,
                        fracking=fracking,
                        frack_step=frack_step,
                        post_burn=post_burn)

    def generate_dummy_data(self,
                            name,
                            max_time=1000.,
                            plot_points=100,
                            band_list=['V'],
                            band_systems=[''],
                            band_instruments=['']):
        time_list = np.linspace(0.0, max_time, plot_points)
        times = np.repeat(time_list, len(band_list))
        bands = [i for s in [band_list for x in time_list] for i in s]
        systs = [i for s in [band_systems for x in time_list] for i in s]
        insts = [i for s in [band_instruments for x in time_list] for i in s]

        data = {name: {'photometry': []}}
        for ti, time in enumerate(times):
            band = bands[ti]
            if isinstance(band, dict):
                band = band['name']

            photodict = {'time': time,
                         'band': band,
                         'magnitude': 0.0,
                         'e_magnitude': 0.0}
            if systs[ti]:
                photodict['system'] = systs[ti]
            if insts[ti]:
                photodict['instrument'] = insts[ti]
            data[name]['photometry'].append(photodict)

        return data



import argparse
import os

from mosfit.fitter import Fitter

from . import __version__


def main():
    """First, parse command line arguments.
    """

    dir_path = os.path.dirname(os.path.realpath(__file__))

    parser = argparse.ArgumentParser(
        prog='MOSFiT',
        description='Fit astrophysical light curves using AstroCats data.')

    parser.add_argument(
        '--events',
        '-e',
        dest='events',
        default=[''],
        nargs='+',
        help=("List of event names to be fit, delimited by spaces. If an "
              "event name contains a space, enclose the event's name in "
              "double quote marks, e.g. \"SDSS-II SN 5944\"."))

    parser.add_argument(
        '--models',
        '-m',
        dest='models',
        default=['default'],
        nargs='+',
        help=("List of models to use to fit against the listed events. The "
              "model can either be a name of a model included with MOSFiT, or "
              "a path to a custom model JSON file generated by the user."))

    parser.add_argument(
        '--parameter-paths',
        '-P',
        dest='parameter_paths',
        default=[''],
        nargs='+',
        help=("Paths to parameter files corresponding to each model file; "
              "length of this list should be equal to the length of the list "
              "of models"))

    parser.add_argument(
        '--plot-points',
        dest='plot_points',
        default=100,
        help=("Set the number of plot points when producing light curves from "
              "models without fitting against any actual transient data."))

    parser.add_argument(
        '--max-time',
        dest='max_time',
        default=1000.,
        help=("Set the maximum time for model light curves to be plotted "
              "until."))

    parser.add_argument(
        '--band-list',
        dest='band_list',
        default=['V'],
        help=("List of bands to plot when plotting model light curves that "
              "are not being matched to actual transient data."))

    parser.add_argument(
        '--band-systems',
        dest='band_systems',
        default=[''],
        help=("List of photometric systems corresponding to the bands listed "
              "in `--band-list`."))

    parser.add_argument(
        '--band-instruments',
        dest='band_instruments',
        default=[''],
        help=("List of instruments corresponding to the bands listed "
              "in `--band-list`."))

    parser.add_argument(
        '--iterations',
        '-i',
        dest='iterations',
        type=int,
        default=1000,
        help=("Number of iterations to run emcee for, including burn-in and "
              "post-burn iterations."))

    parser.add_argument(
        '--num-walkers',
        '-N',
        dest='num_walkers',
        type=int,
        default=50,
        help=("Number of walkers to use in emcee, must be at least twice the "
              "total number of free parameters within the model."))

    parser.add_argument(
        '--num-temps',
        '-T',
        dest='num_temps',
        type=int,
        default=2,
        help=("Number of temperatures to use in the parallel-tempered emcee "
              "sampler. `-T 1` is equivalent to the standard "
              "EnsembleSampler."))

    parser.add_argument(
        '--no-fracking',
        dest='fracking',
        default=True,
        action='store_false',
        help=("Setting this flag will skip the `fracking` step of the "
              "optimization process."))

    parser.add_argument(
        '--frack-step',
        '-f',
        dest='frack_step',
        type=int,
        default=20,
        help=("Perform `fracking` every this number of steps while in the "
              "burn-in phase of the fitting process."))

    parser.add_argument(
        '--post-burn',
        '-p',
        dest='post_burn',
        type=int,
        default=500,
        help=("Run emcee this many more iterations after the burn-in phase. "
              "The burn-in phase will thus be run for (i - p) iterations, "
              "where i is the total number of iterations set with `-i` and "
              "p is the value of this parameter."))

    parser.add_argument(
        '--travis',
        dest='travis',
        default=False,
        action='store_true',
        help=("Alters the printing of output messages such that a new line is "
              "generated with each message. Users are unlikely to need this "
              "parameter; it is included as Travis requires new lines to be "
              "produed to detected program output."))

    args = parser.parse_args()

    # Print our amazing ASCII logo.
    with open(os.path.join(dir_path, 'logo.txt'), 'r') as f:
        logo = f.read()
        width = len(logo.split('\n')[0])
        aligns = '{:^' + str(width) + '}'
        print(logo)
    print((aligns + '\n').format('### MOSFiT -- version {} ###'.format(
        __version__)))
    print(aligns.format('Authored by James Guillochon & Matt Nicholl'))
    print(aligns.format('Released under the MIT license'))
    print((aligns + '\n').format('https://github.com/guillochon/MOSFiT'))

    # Then, fit the listed events with the listed models.
    fitargs = {
        'events': args.events,
        'models': args.models,
        'plot_points': args.plot_points,
        'max_time': args.max_time,
        'band_list': args.band_list,
        'band_systems': args.band_systems,
        'band_instruments': args.band_instruments,
        'iterations': args.iterations,
        'num_walkers': args.num_walkers,
        'num_temps': args.num_temps,
        'parameter_paths': args.parameter_paths,
        'fracking': args.fracking,
        'frack_step': args.frack_step,
        'travis': args.travis,
        'post_burn': args.post_burn
    }
    Fitter().fit_events(**fitargs)


if __name__ == "__main__":
    main()



from math import isnan

import numpy as np

from .engine import Engine

CLASS_NAME = 'Magnetar'


class Magnetar(Engine):
    """Magnetar spin-down engine
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process(self, **kwargs):
        if 'densetimes' in kwargs:
            self._times = kwargs['densetimes']
        else:
            self._times = kwargs['times']
        self._Pspin = kwargs['Pspin']
        self._Bfield = kwargs['Bfield']
        self._Mns = kwargs['Mns']
        self._thetaPB = kwargs['thetaPB']
        self._texplosion = kwargs['texplosion']

        Ep = 2.6e52 * (self._Mns / 1.4)**(3. / 2.) * self._Pspin**(-2)

        tp = 1.3e5 * self._Bfield**(-2) * self._Pspin**2 * (self._Mns / 1.4)**(
            3. / 2.) * (np.sin(self._thetaPB))**(-2)

        ts = [np.inf if self._texplosion > x else (x - self._texplosion)
              for x in self._times]

        luminosities = [Ep / tp / (1. + t / tp)**2 for t in ts]
        luminosities = [0.0 if isnan(x) else x for x in luminosities]

        # Add on to any existing luminosity
        luminosities = self.add_to_existing_lums(luminosities)

        return {'luminosities': luminosities}



from ..module import Module

CLASS_NAME = 'Engine'


class Engine(Module):
    """Generic engine module.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process(self, **kwargs):
        return {}

    def add_to_existing_lums(self, new_lums, **kwargs):
        # Add on to any existing luminosity
        old_lums = kwargs.get('luminosities', None)
        if old_lums is not None:
            new_lums = [x + y for x, y in zip(old_lums, new_lums)]
        return new_lums



from math import isnan

import numpy as np

from ...constants import LIKELIHOOD_FLOOR
from ..module import Module

CLASS_NAME = 'Likelihood'


class Likelihood(Module):
    """Calculate the maximum likelihood score for a model.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process(self, **kwargs):
        self._model_mags = kwargs['model_magnitudes']
        self._fractions = kwargs['fractions']
        for mag in self._model_mags:
            if isnan(mag):
                return {'value': LIKELIHOOD_FLOOR}
        self._variance2 = kwargs['variance']**2
        self._mags = kwargs['magnitudes']
        self._e_mags = kwargs['e_magnitudes']
        self._n_mags = len(self._mags)

        value = -0.5 * np.sum(
            [(x - y)**2 /
             (z**2 + self._variance2) + np.log(self._variance2 + z**2)
             for x, y, z in zip(self._model_mags, self._mags, self._e_mags)])
        if isnan(value):
            return {'value': LIKELIHOOD_FLOOR}

        x = self._fractions
        if min(x) < 0.0 or max(x) > 1.0:
            value = LIKELIHOOD_FLOOR
        # if min(x) < 0.0:
        #     value = value + self._n_mags * np.sum([y for y in x if y < 0.0])
        # if max(x) > 1.0:
        #     value = value + self._n_mags * np.sum(
        #         [1.0 - y for y in x if y > 1.0])
        return {'value': value}



from ...constants import DAY_CGS
from ..module import Module

CLASS_NAME = 'Transform'


class Transform(Module):
    """Parent class for transforms.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process(self, **kwargs):
        return {}

    def set_times_lums(self, **kwargs):
        self._times = kwargs['times']
        self._t_explosion = kwargs['texplosion']
        if 'densetimes' in kwargs:
            self._dense_times = kwargs['densetimes']
            self._luminosities = kwargs['luminosities']
        elif min(self._times) > self._t_explosion:
            self._dense_times = [self._t_explosion] + kwargs['times']
            self._luminosities = [0.0] + kwargs['luminosities']
        self._times_since_exp = [(x - self._t_explosion) * DAY_CGS
                                 for x in self._times]
        self._dense_times_since_exp = [(x - self._t_explosion) * DAY_CGS
                                       for x in self._dense_times]



from .sed import SED

CLASS_NAME = 'Line'


class Line(SED):
    """Line spectral energy distribution, modifies existing SED.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process(self, **kwargs):
        self._seds = kwargs['seds']
        self._band_wavelengths = kwargs['bandwavelengths']
        self._luminosities = kwargs['luminosities']
        zp1 = 1.0 + kwargs['redshift']

        seds = []
        for li, lum in enumerate(self._luminosities):
            cur_band = self._bands[li]
            bi = self._filters.find_band_index(cur_band)
            rest_freqs = [x * zp1 for x in self._band_frequencies[bi]]

            # Dummy function for now, needs implementation
            sed = [0.0 for x in rest_freqs]
            seds.append(list(sed))

        seds = self.add_to_existing_seds(seds, **kwargs)

        return {'bandwavelengths': self._band_wavelengths, 'seds': seds}



import numpy as np

from ..module import Module

CLASS_NAME = 'DenseTimes'


class DenseTimes(Module):
    """This class ensures an even time-sampling between the time of explosion
    and the last datapoint, as many transients may lack regular candence data.
    """

    N_TIMES = 100

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._n_times = kwargs[
            'n_times'] if 'n_times' in kwargs else self.N_TIMES

    def process(self, **kwargs):
        self._times = kwargs['times']
        self._t_explosion = kwargs['texplosion']

        outputs = {}
        max_times = max(kwargs['times'])
        if max_times > kwargs['texplosion']:
            outputs['densetimes'] = sorted([
                x + self._t_explosion
                for x in np.linspace(
                    0.0, max_times - self._t_explosion, num=self._n_times)
            ] + self._times)
        else:
            outputs['densetimes'] = kwargs['times']
        return outputs



from math import pi

import numexpr as ne
import numpy as np
from astropy import constants as c

from ...constants import DAY_CGS, FOUR_PI, KM_CGS, M_SUN_CGS
from .sed import SED

CLASS_NAME = 'EnvelopeCore'


class EnvelopeCore(SED):
    """Expanding/receding photosphere with a core+envelope
    morphology and a blackbody spectral energy
    distribution.
    """

    FLUX_CONST = FOUR_PI * (2.0 * c.h / (c.c**2) * pi).cgs.value
    X_CONST = (c.h / c.k_B).cgs.value
    STEF_CONST = (4.0 * pi * c.sigma_sb).cgs.value
    PL_ENV = 10.0

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process(self, **kwargs):
        self._t_explosion = kwargs['texplosion']
        self._times = kwargs['times']
        self._luminosities = kwargs['luminosities']
        self._temperature = kwargs['temperature']
        self._bands = kwargs['bands']
        self._v_ejecta = kwargs['vejecta']
        self._m_ejecta = kwargs['mejecta']
        self._kappa = kwargs['kappa']
        xc = self.X_CONST
        fc = self.FLUX_CONST
        slope = self.PL_ENV
        zp1 = 1.0 + kwargs['redshift']
        seds = []
        for li, lum in enumerate(self._luminosities):
            cur_band = self._bands[li]
            bi = self._filters.find_band_index(cur_band)
            rest_freqs = [x * zp1 for x in self._band_frequencies[bi]]
            # rest_freqs3 = [x**3 for x in rest_freqs]

            # Radius is determined via expansion, unless this would make
            # temperature lower than temperature parameter.
            radius = self._v_ejecta * KM_CGS * (
                self._times[li] - self._t_explosion) * DAY_CGS

            # Compute density in core
            rho_core = (3.0 * self._m_ejecta * M_SUN_CGS /
                        (4.0 * pi * radius**3))

            # Attach power-law envelope of negligible mass
            tau_e = self._kappa * rho_core * radius / (slope - 1.0)

            # Find location of photosphere in envelope/core
            if tau_e > 0.667:
                radius_phot = (2.0 * (slope - 1.0) /
                               (3.0 * self._kappa * rho_core * radius
                                **slope))**(1.0 / (1.0 - slope))
            else:
                radius_phot = slope * radius / (slope - 1.0) - 2.0 / (
                    3.0 * self._kappa * rho_core)

            # Put temperature floor to prevent weird behaviour as r_phot -> 0
            rec_radius = np.sqrt(lum /
                                 (self.STEF_CONST * self._temperature**4))
            if radius < rec_radius:
                radius2 = radius**2
                temperature = (lum / (self.STEF_CONST * radius2))**0.25
            else:
                radius2 = rec_radius**2
                temperature = self._temperature

            if li == 0:
                sed = ne.evaluate('fc * radius2 * rest_freqs**3 / '
                                  'exp(xc * rest_freqs / temperature) - 1.0')
            else:
                sed = ne.re_evaluate()
            # a = [np.exp(self.X_CONST * x / temperature) - 1.0
            #      for x in rest_freqs]
            # sed = [
            #     (self.FLUX_CONST * radius2 * x / y)
            #     for x, y in zip(rest_freqs3, a)
            # ]
            seds.append(list(sed))

        seds = self.add_to_existing_seds(seds, **kwargs)

        return {'bandwavelengths': self._band_wavelengths, 'seds': seds}



import numpy as np

from ..module import Module

CLASS_NAME = 'Parameter'


class Parameter(Module):
    """Model parameter that can either be free or fixed.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._max_value = kwargs.get('max_value', None)
        self._min_value = kwargs.get('min_value', None)
        self._value = kwargs.get('value', None)
        self._log = kwargs.get('log', False)
        self._latex = kwargs.get('latex', self._name)
        if (self._log and self._min_value is not None and
                self._max_value is not None):
            self._min_value = np.log(self._min_value)
            self._max_value = np.log(self._max_value)

    def latex(self):
        return self._latex

    def process(self, **kwargs):
        """Initialize a parameter based upon either a fixed value or a
        distribution, if one is defined.
        """

        if self._min_value is None or self._max_value is None:
            # If this parameter is not free and is already set, then skip
            if self._name in kwargs:
                return {}

            value = self._value
        else:
            value = max(
                min((kwargs['fraction'] *
                     (self._max_value - self._min_value) + self._min_value),
                    self._max_value), self._min_value)
            if self._log:
                value = np.exp(value)
        return {self._name: value}



from setuptools import setup, find_packages

exec(open('mosfit/__init__.py').read())

setup(
    name='mosfit',
    packages=find_packages(),
    include_package_data=True,
    version=__version__,
    description=('Package that performs maximum likelihood analysis to fit '
                 'semi-analytical model predictions to observed '
                 'astronomical transient data.'),
    license=__license__,
    author=__author__,
    author_email='guillochon@gmail.com',
    url='https://github.com/guillochon/mosfit',
    download_url='https://github.com/guillochon/mosfit/tarball/0.1.1',
    keywords=['astronomy', 'fitting', 'monte carlo', 'modeling'],
    classifiers=['Development Status :: 3 - Alpha',
                 'License :: OSI Approved :: MIT License',
                 'Natural Language :: English',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3 :: Only',
                 'Topic :: Scientific/Engineering :: Astronomy',
                 'Topic :: Scientific/Engineering :: Physics'])



from math import isnan

import numexpr as ne
import numpy as np

from ...constants import C_CGS, DAY_CGS, FOUR_PI, KM_CGS, M_SUN_CGS
from .transform import Transform

CLASS_NAME = 'Diffusion'


class Diffusion(Transform):
    """Photon diffusion transform.
    """

    N_INT_TIMES = 20
    MIN_EXP_ARG = 20
    DIFF_CONST = 2.0 * M_SUN_CGS / (13.7 * C_CGS * KM_CGS)
    TRAP_CONST = 3.0 * M_SUN_CGS / (FOUR_PI * KM_CGS**2)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def process(self, **kwargs):
        self.set_times_lums(**kwargs)
        self._kappa = kwargs['kappa']
        self._kappa_gamma = kwargs['kappagamma']
        self._m_ejecta = kwargs['mejecta']
        self._v_ejecta = kwargs['vejecta']
        self._tau_diff = np.sqrt(self.DIFF_CONST * self._kappa *
                                 self._m_ejecta / self._v_ejecta)
        self._trap_coeff = (self.TRAP_CONST * self._kappa_gamma *
                            self._m_ejecta / (self._v_ejecta**2))
        td2, A = self._tau_diff**2, self._trap_coeff
        tbarg = self.MIN_EXP_ARG * td2

        new_lum = []
        evaled = False
        lum_cache = {}
        min_te = min(self._dense_times_since_exp)
        for te in self._times_since_exp:
            if te <= 0.0:
                new_lum.append(0.0)
                continue
            if te in lum_cache:
                new_lum.append(lum_cache[te])
                continue
            te2 = te**2
            tb = max(np.sqrt(max(te2 - tbarg, 0.0)), min_te)
            int_times = np.linspace(tb, te, self.N_INT_TIMES)
            dt = int_times[1] - int_times[0]

            int_lums = np.interp(int_times, self._dense_times_since_exp,
                                 self._luminosities)

            if not evaled:
                int_arg = ne.evaluate('2.0 * int_lums * int_times / td2 * '
                                      'exp((int_times**2 - te2) / td2) * '
                                      '(1.0 - exp(-A / te2))')
                evaled = True
            else:
                int_arg = ne.re_evaluate()
            # int_arg = [
            #     2.0 * l * t / td2 *
            #     np.exp((t**2 - te**2) / td2) * (1.0 - np.exp(-A / te**2))
            #     for t, l in zip(int_times, int_lums)
            # ]
            int_arg = [0.0 if isnan(x) else x for x in int_arg]
            lum_val = np.trapz(int_arg, dx=dt)
            lum_cache[te] = lum_val
            new_lum.append(lum_val)
        return {'luminosities': new_lum}
